! Definition and allocation of parallel datas for OpenMP.

!-
!- $Header: /home/ssipsl/CVSREP/ORCHIDEE/src_parallel/Attic/mod_orchidee_omp_data.F90,v 1.1.2.3 2008/11/05 10:08:25 ssipsl Exp $
!- 

MODULE mod_orchidee_omp_data

!-
  USE defprec
  USE ioipsl

  IMPLICIT NONE

  ! Check all OpenMP transferts in ORCHIDEE : use this flag to debug synchronism with OpenMP 
  LOGICAL, PARAMETER :: check_all_transfert=.FALSE.


  INTEGER(i_std),SAVE :: omp_size
  INTEGER(i_std),SAVE :: omp_rank
  LOGICAL,SAVE :: is_omp_root
  LOGICAL,SAVE :: is_ok_omp
!$OMP  THREADPRIVATE(omp_rank,is_omp_root)
  
  INTEGER(i_std),SAVE,DIMENSION(:),ALLOCATABLE :: nbp_omp_para_nb
  INTEGER(i_std),SAVE,DIMENSION(:),ALLOCATABLE :: nbp_omp_para_begin
  INTEGER(i_std),SAVE,DIMENSION(:),ALLOCATABLE :: nbp_omp_para_end    
  
  INTEGER(i_std),SAVE :: nbp_omp
  INTEGER(i_std),SAVE :: nbp_omp_begin
  INTEGER(i_std),SAVE :: nbp_omp_end
  INTEGER(i_std),SAVE :: offset_omp
!$OMP  THREADPRIVATE(nbp_omp,nbp_omp_begin,nbp_omp_end,offset_omp)

  ! Flag for each OMP process for ORCHIDEE to verify synchronization if function Synchro_Omp is used
  LOGICAL,SAVE,ALLOCATABLE :: proc_synchro_omp(:)

  INTEGER, SAVE :: numout_omp = -1
!$OMP  THREADPRIVATE(numout_omp)

  ! For debugging OpenMP processes : id of OMP function for each task.
  ! If one task is not in the same function, we can see it.
  INTEGER(i_std),SAVE,ALLOCATABLE,DIMENSION(:) :: omp_function
  ! It is not SHARED.
  ! List of values :
  CHARACTER(LEN=28), PARAMETER :: omp_fct_name(-1:72) = (/ &
           "Initialization              ", &
           "Synchro_Omp                 ", &
           "check_buffer_i              ", &
           "check_buffer_r              ", &
           "check_buffer_l              ", &
           "bcast_omp_c                 ", &
           "bcast_omp_i                 ", &
           "bcast_omp_i1                ", &
           "bcast_omp_i2                ", &
           "bcast_omp_i3                ", &
           "bcast_omp_i4                ", &
           "bcast_omp_r                 ", &
           "bcast_omp_r1                ", &
           "bcast_omp_r2                ", &
           "bcast_omp_r3                ", &
           "bcast_omp_r4                ", &
           "bcast_omp_l                 ", &
           "bcast_omp_l1                ", &
           "bcast_omp_l2                ", &
           "bcast_omp_l3                ", &
           "bcast_omp_l4                ", &
           "scatter_omp_i               ", &
           "scatter_omp_i1              ", &
           "scatter_omp_i2              ", &
           "scatter_omp_i3              ", &
           "scatter_omp_r               ", &
           "scatter_omp_r1              ", &
           "scatter_omp_r2              ", &
           "scatter_omp_r3              ", &
           "scatter_omp_l               ", &
           "scatter_omp_l1              ", &
           "scatter_omp_l2              ", &
           "scatter_omp_l3              ", &
           "gather_omp_i0               ", &
           "gather_omp_i                ", &
           "gather_omp_i1               ", &
           "gather_omp_i2               ", &
           "gather_omp_i3               ", &
           "gather_omp_r0               ", &
           "gather_omp_r                ", &
           "gather_omp_r1               ", &
           "gather_omp_r2               ", &
           "gather_omp_r3               ", &
           "gather_omp_l0               ", &
           "gather_omp_l                ", &
           "gather_omp_l1               ", &
           "gather_omp_l2               ", &
           "gather_omp_l3               ", &
           "reduce_sum_omp_i            ", &
           "reduce_sum_omp_i1           ", &
           "reduce_sum_omp_i2           ", &
           "reduce_sum_omp_i3           ", &
           "reduce_sum_omp_i4           ", &
           "reduce_sum_omp_r            ", &
           "reduce_sum_omp_r1           ", &
           "reduce_sum_omp_r2           ", &
           "reduce_sum_omp_r3           ", &
           "reduce_sum_omp_r4           ", &
           "orch_bcast_omp_cgen         ", &
           "orch_bcast_omp_igen         ", &
           "orch_bcast_omp_rgen         ", &
           "orch_bcast_omp_lgen         ", &
           "orch_scatter_omp_igen       ", &
           "orch_scatter_omp_rgen       ", &
           "orch_scatter_omp_lgen       ", &
           "orch_gather_omp_simple_igen ", &
           "orch_gather_omp_igen        ", &
           "orch_gather_omp_simple_rgen ", &
           "orch_gather_omp_rgen        ", &
           "orch_gather_omp_simple_lgen ", &
           "orch_gather_omp_lgen        ", &
           "orch_reduce_sum_omp_igen    ", &
           "orch_reduce_sum_omp_rgen    ", &
           "check_buffer_c              " /)

  ! Previous value for own omp_function
  INTEGER, SAVE :: omp_previous
!$OMP  THREADPRIVATE(omp_previous)

CONTAINS

    SUBROUTINE barrier2_omp()

    IMPLICIT NONE

!$OMP BARRIER
!$OMP BARRIER

  END SUBROUTINE barrier2_omp



  
  SUBROUTINE Init_orchidee_omp
  IMPLICIT NONE
  
#ifdef CPP_OMP
    IF (is_omp_root) THEN
        is_ok_omp=.TRUE.
    ENDIF
#else    
    is_ok_omp=.FALSE.
#endif


    IF (is_ok_omp) THEN
      STOP 'Open MP is not yet implemented for driver'
    ELSE
      omp_size=1
      omp_rank=0
      is_omp_root=.TRUE.
    ENDIF

  END SUBROUTINE Init_orchidee_omp

  SUBROUTINE Init_numout_omp(numout)
    INTEGER, INTENT(in) :: numout
    numout_omp=numout
  END SUBROUTINE Init_numout_omp


  SUBROUTINE Init_orchidee_omp_data(arg_omp_size,arg_omp_rank,arg_nbp_omp,arg_offset_omp)
    IMPLICIT NONE
    INTEGER, INTENT(IN) :: arg_omp_size
    INTEGER, INTENT(IN) :: arg_omp_rank
    INTEGER, INTENT(IN) :: arg_nbp_omp
    INTEGER, INTENT(IN) :: arg_offset_omp
    
    INTEGER    :: i
    
    
    IF (arg_omp_rank==0) THEN
      is_omp_root=.TRUE.
    ELSE
      is_omp_root=.FALSE.
    ENDIF
    
#ifdef CPP_OMP
    IF (is_omp_root) THEN
        is_ok_omp=.TRUE.
    ENDIF
#else    
    is_ok_omp=.FALSE.
#endif

    IF (is_omp_root) omp_size=arg_omp_size

    CALL barrier2_omp()

     omp_rank=arg_omp_rank
    
    IF (is_omp_root) THEN 
      ALLOCATE(nbp_omp_para_nb(0:omp_size-1))
      ALLOCATE(nbp_omp_para_begin(0:omp_size-1))
      ALLOCATE(nbp_omp_para_end(0:omp_size-1))
    ENDIF
    
    CALL barrier2_omp()
    nbp_omp_para_nb(omp_rank)=arg_nbp_omp
    CALL barrier2_omp()
    
    IF (is_omp_root) THEN

      nbp_omp_para_begin(0)=1
      nbp_omp_para_end(0)=nbp_omp_para_nb(0)

      DO i=1,omp_size-1
        nbp_omp_para_begin(i)=nbp_omp_para_end(i-1)+1
        nbp_omp_para_end(i)=nbp_omp_para_begin(i)+nbp_omp_para_nb(i)-1
      ENDDO

    ENDIF

    CALL barrier2_omp()
     
    nbp_omp=nbp_omp_para_nb(omp_rank)
    nbp_omp_begin=nbp_omp_para_begin(omp_rank)
    nbp_omp_end=nbp_omp_para_end(omp_rank)
    
    offset_omp=arg_offset_omp          
    CALL Print_omp_data
    
    CALL Init_synchro_omp()
    
  END SUBROUTINE Init_orchidee_omp_data

  SUBROUTINE print_omp_data
  IMPLICIT NONE

!$OMP CRITICAL  
  PRINT *,'--------> ORCHIDEE TASK ',omp_rank
  PRINT *,'omp_size =',omp_size
  PRINT *,'omp_rank =',omp_rank
  PRINT *,'is_omp_root =',is_omp_root
  PRINT *,'offset_omp',offset_omp
  PRINT *,'nbp_omp_para_nb =',nbp_omp_para_nb
  PRINT *,'nbp_omp_para_begin =',nbp_omp_para_begin
  PRINT *,'nbp_omp_para_end =',nbp_omp_para_end    
  PRINT *,'nbp_omp =',nbp_omp
  PRINT *,'nbp_omp_begin =',nbp_omp_begin
  PRINT *,'nbp_omp_end =',nbp_omp_end    
!$OMP END CRITICAL

  END SUBROUTINE print_omp_data

  SUBROUTINE Init_synchro_omp
  IMPLICIT NONE
    
    IF (is_omp_root) THEN
      ALLOCATE(proc_synchro_omp(0:omp_size-1))
      proc_synchro_omp(:)=.FALSE.

      IF ( check_all_transfert ) THEN
         ALLOCATE(omp_function(0:omp_size-1))
         omp_function(:)=-1
      ENDIF
    ENDIF
    CALL barrier2_omp()

  END SUBROUTINE Init_Synchro_omp
  
  SUBROUTINE Synchro_omp
  IMPLICIT NONE
    INTEGER iter
    LOGICAL, PARAMETER :: check=.TRUE.
    INTEGER, PARAMETER :: iter_max=1
    INTEGER, PARAMETER :: print_iter=1

    proc_synchro_omp(omp_rank)=.TRUE.
    CALL barrier2_omp()

    iter=0
    DO WHILE (.NOT. ALL(proc_synchro_omp))
       iter=iter+1
       IF ( mod(iter,print_iter) == 0 ) THEN
          IF (numout_omp > 0) THEN
             WRITE(numout_omp,*) "ORCHIDEE SYNCHRO OMP : iter ",iter," rank ",omp_rank," wait for ",proc_synchro_omp
          ELSE
             WRITE(*,*) "ORCHIDEE SYNCHRO OMP : iter ",iter," rank ",omp_rank," wait for ",proc_synchro_omp
          ENDIF
       ENDIF
       IF (check) THEN
          IF (iter > iter_max) THEN
             IF (numout_omp > 0) THEN
                WRITE(numout_omp,*) "TOO MUCH WAIT in Synchro_Omp !! iter ",iter," rank ",omp_rank," wait for ",proc_synchro_omp
                WRITE(numout_omp,*) "We stop here"
                WRITE(numout_omp,*) "omp_function : ",omp_function(:)
             ELSE
                WRITE(*,*) "TOO MUCH WAIT in Synchro_Omp !! iter ",iter," rank ",omp_rank," wait for ",proc_synchro_omp
                WRITE(*,*) "We stop here"
                WRITE(*,*) "omp_function : ",omp_function(:)
             ENDIF
#ifdef CPP_PARA
             CALL MPI_ABORT(4)
#endif     
             STOP 'Fatal error from ORCHIDEE : Synchro_Omp failed'
          ENDIF
       ENDIF
    CALL barrier2_omp()
    ENDDO
    CALL barrier2_omp()
    proc_synchro_omp(omp_rank)=.FALSE.
    CALL barrier2_omp()

   END SUBROUTINE Synchro_omp

   SUBROUTINE print_omp_function ()

     IF ( check_all_transfert ) THEN
        CALL barrier2_omp()
        IF (numout_omp > 0) THEN
           WRITE(numout_omp,*) omp_rank,&
                " : ",omp_fct_name(omp_previous),'->',omp_fct_name(omp_function(omp_rank))
           IF (MINVAL(omp_function(:)).LT.MAXVAL(omp_function(:))) &
                WRITE(numout_omp,*) "!!! OMP ERROR : NO MORE SYNCHRO  !!!  ",omp_function(:)
        ELSE
           WRITE(*,*) omp_rank,&
                " : ",omp_fct_name(omp_previous),'->',omp_fct_name(omp_function(omp_rank))
           IF (MINVAL(omp_function(:)).LT.MAXVAL(omp_function(:))) &
                WRITE(*,*) "!!! OMP ERROR : NO MORE SYNCHRO  !!!  ",omp_function(:)
        ENDIF
        CALL barrier2_omp()
     ENDIF

  END SUBROUTINE print_omp_function


END MODULE mod_orchidee_omp_data
